package fos

import scala.collection.mutable.{Map,HashMap}

case class TypeError(msg: String) extends Exception(msg)

object Type {
  import CT._
  import Utils._
  
  type Class = String
  type Context = List[(Class, String)]
  
  def typeOf(tree: Tree, ctx: Context): Class = null // TODO: implement
}

case class EvaluationException(msg: String) extends Exception

object Evaluate extends (Expr => Expr) {
  import Utils._
  
  def substituteInBody(exp: Expr, thiss: New, substs: List[(FieldDef,Expr)]): Expr = exp match {
    case Select(obj: Expr, field: String) => Select(substituteInBody(obj, thiss, substs), field)
    case New(cls, args) => New(cls, args map (arg => substituteInBody(arg, thiss, substs)))
    case Cast(cls, e) => Cast(cls, substituteInBody(e, thiss, substs))
    case Var("this") => thiss
    case Var(bd) => substs find (_._1.name == bd) match {
      case None => exp
      case Some((_,sub)) => sub
    }
    case Apply(obj,method,args) => Apply(substituteInBody(obj, thiss, substs), method, args map { e => substituteInBody(e, thiss, substs) } )
    case _ => throw EvaluationException("Apply: Forgot expression "+exp)
  }
  
  def apply(expr: Expr): Expr = null // TODO: implement
}

object CT {
  val objectClass = "Object"
  private val objectClassDef = ClassDef(objectClass, null, Nil, CtrDef(objectClass, Nil, Nil, Nil) , Nil)
  private val ct = new HashMap[String, ClassDef]
   
  add(objectClass,objectClassDef)
  
  def elements = ct.iterator
  def lookup(classname: String): Option[ClassDef] = if(classname != null) ct get classname else None
  def add(key: String, element: ClassDef): Unit = ct += key -> element
  def delete(key: String) = ct -= key
  def clear(): Unit = {
    ct.clear
    add(objectClass,objectClassDef)
  }
}


object Utils {
  def getClassDef(className: String): ClassDef = CT lookup className match {
    case None => throw TypeError("class "+className+" not declared")
    case Some(c: ClassDef) => c
  }
}